//   Copyright 2007, 2008 Nicolas Dubuit
//
//     This file is part of Immix.
//
//     Immix is free software: you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation, either version 3 of the License, or
//     (at your option) any later version.
//
//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "correlator.h"

#include <stdio.h>
#include <iostream>
#include <cmath>
using namespace std ;

QMutex Correlator::fftwMutex ;

Correlator::Correlator (int sz, int fctr)
{
    ImgSz = sz ;
    FineFactor = fctr ;
    CorSz = FineFactor*ImgSz * 2 ;
    
    fftwMutex.lock() ;

    refGray = (fftw_complex*) fftw_malloc(CorSz*CorSz * sizeof(fftw_complex)) ;
    targGray = (fftw_complex*) fftw_malloc(CorSz*CorSz * sizeof(fftw_complex)) ;
    
    fRefGray = (fftw_complex*) fftw_malloc(CorSz*CorSz * sizeof(fftw_complex)) ;
    fTargGray = (fftw_complex*) fftw_malloc(CorSz*CorSz * sizeof(fftw_complex)) ;
    
    Correlation = targGray ;
///    Correlation = (fftw_complex*) fftw_malloc(CorSz*CorSz * sizeof(fftw_complex)) ;

    pTargGray = fftw_plan_dft_2d(CorSz, CorSz, targGray, fTargGray, FFTW_FORWARD, FFTW_ESTIMATE);
    pRefGray = fftw_plan_dft_2d(CorSz, CorSz, refGray, fRefGray, FFTW_FORWARD, FFTW_ESTIMATE);
    pBackTarg = fftw_plan_dft_2d(CorSz, CorSz, fTargGray, Correlation, FFTW_BACKWARD, FFTW_ESTIMATE);

    fftwMutex.unlock() ;
}


Correlator::~Correlator (void)
{
    fftwMutex.lock() ;

    // Clean up FFTW suff.
	fftw_destroy_plan(pTargGray);
	fftw_destroy_plan(pRefGray);
	fftw_destroy_plan(pBackTarg);
	
	// Release FFT arrays.
    if (targGray)
        fftw_free(targGray); 
    if (refGray)
       	fftw_free(refGray);
	if (fTargGray)
        fftw_free(fTargGray);
    if (fRefGray)
        fftw_free(fRefGray);

    fftwMutex.unlock() ;
}


void Correlator::set_data (QImage& img, fftw_complex* data)
{
	QImage scaledImg = img.scaled (FineFactor*ImgSz,FineFactor*ImgSz,Qt::IgnoreAspectRatio, Qt::SmoothTransformation ) ;
	for (int x=0 ; x<CorSz ; x++ ) 
	  for (int y=0 ; y<CorSz ; y++ )
	  {
        if (x>CorSz/4 && x<CorSz/4*3 && y>CorSz/4 && y<CorSz/4*3)
        {
            data[y+CorSz*x][0] = qGray(scaledImg.pixel(x-CorSz/4,y-CorSz/4)) ;
            data[y+CorSz*x][1] = 0. ;
        }
        else
        {
            data[y+CorSz*x][0] = 0. ;
            data[y+CorSz*x][1] = 0. ;
        }
	  }
}

void Correlator::set_data_square (QImage& img, fftw_complex* data)
{
	QImage scaledImg = img.scaled (FineFactor*ImgSz,FineFactor*ImgSz,Qt::IgnoreAspectRatio, Qt::SmoothTransformation ) ;
    int quarterSz = CorSz/4 ;
	for (int x=0 ; x<CorSz ; x++ ) 
	  for (int y=0 ; y<CorSz ; y++ )
	  {
        if (x>quarterSz && x<3*quarterSz && y>quarterSz && y<3*quarterSz)
        {
            data[y+CorSz*x][0] =   qGray(scaledImg.pixel(x-quarterSz,y-quarterSz))
                                 * qGray(scaledImg.pixel(x-quarterSz,y-quarterSz)) ;
            data[y+CorSz*x][1] = 0. ;
        }
        else
        {
            data[y+CorSz*x][0] = 0. ;
            data[y+CorSz*x][1] = 0. ;
        }
	  }
}

void Correlator::set_window_data (fftw_complex* data)
{
    int quarterSz = CorSz/4 ;
	for (int x=0 ; x<CorSz ; x++ ) 
	  for (int y=0 ; y<CorSz ; y++ )
	  {
        if (x>quarterSz && x<3*quarterSz && y>quarterSz && y<3*quarterSz)
        {
            data[y+CorSz*x][0] = 1. ;
            data[y+CorSz*x][1] = 0. ;
        }
        else
        {
            data[y+CorSz*x][0] = 0. ;
            data[y+CorSz*x][1] = 0. ;
        }
	  }
}

void Correlator::setRefData (QImage& img) 
{
    if ( ! refGray ) return ;
    set_data (img, refGray) ;
	fftw_execute (pRefGray) ;
}

void Correlator::setTargetData (QImage& img) 
{
    if ( ! targGray ) return ;
    set_data (img, targGray) ;
	fftw_execute (pTargGray) ;
}

void Correlator::setRefDataSquare (QImage& img) 
{
    if ( ! refGray ) return ;
    set_data_square (img, refGray) ;
	fftw_execute (pRefGray) ;
}

void Correlator::setTargetDataSquare (QImage& img) 
{
    if ( ! targGray ) return ;
    set_data_square (img, targGray) ;
	fftw_execute (pTargGray) ;
}

void Correlator::setWindowRefData (void) 
{
    if ( ! refGray ) return ;
    set_window_data (refGray) ;
	fftw_execute (pRefGray) ;
}

void Correlator::setWindowTargetData (void) 
{
    if ( ! targGray ) return ;
    set_window_data (targGray) ;
	fftw_execute (pTargGray) ;
}

void Correlator::calculateCorrelation (void)
{
    if ( ! Correlation || ! fRefGray || ! fTargGray ) 
        return ;

	// Multiply fourier of target by complex cojugate of fourier of reference.
	//  For u=a+ib and v=c+id,  uv* = (a+ib)(c-id) = ac+bd + i(bc-ad)
	double targr, targi ;
    int x,y ;
	for (x=0 ; x<CorSz ; x++ ) 
	{
	  for (y=0 ; y<CorSz ; y++ )
	  {
 		int pos = y + CorSz * x;
	  	// Real part.
		targr =  ( fRefGray[pos][0] * fTargGray[pos][0] )
		        + ( fRefGray[pos][1] * fTargGray[pos][1] ) ;
        // Imaginary part.
		targi =  ( fRefGray[pos][1] * fTargGray[pos][0])
	            - ( fRefGray[pos][0] * fTargGray[pos][1] ) ;
		fTargGray[pos][0] =  targr ;
		fTargGray[pos][1] =  targi ;
	  }
    }
		
	// Unfourierize the result ; we get the corelation fctn.
	fftw_execute (pBackTarg) ;
}

void Correlator::findCorrelMax (double& dXmax, double& dYmax, 
                        double& corMin, double& corMax, double& corIMin, double& corIMax)
{
    if ( ! Correlation ) 
        return ;

	corMax = Correlation[0][0] ;
	int dxCorMaxInt = 0, dyCorMaxInt = 0 ;

	corMin = Correlation[0][0];
	corIMin = Correlation[1][0] ;
	corIMax = Correlation[1][0] ;

// Main loop (over all pixels). This one should be unrolled (sequential access to data)
	int x,y ;
    for (x=0 ; x<CorSz ; x++ )
	{
	  for (y=0 ; y<CorSz ; y++ )
	  {
 		int pos = y + CorSz * x;
	  	if ( Correlation[pos][0] > corMax )
	  	{
	  		corMax = Correlation[pos][0] ;
	  		dxCorMaxInt = x ;
	  		dyCorMaxInt = y ;
  		}
	  	if ( Correlation[pos][0] < corMin )
	  		corMin = Correlation[pos][0] ;
	  	if ( Correlation[pos][1] < corIMin )
	  		corIMin = Correlation[pos][1] ;
	  	if ( Correlation[pos][1] > corIMax )
	  		corIMax = Correlation[pos][1] ;
  	  }
	}

// If the maximum is in the second half of the correlation, is is meant 
//   to be in the negative values. eg. k=-2 is at CorSz-2... 
	dxCorMaxInt = - ( (dxCorMaxInt + CorSz/2) % (CorSz) - CorSz/2 );
	dyCorMaxInt = - ( (dyCorMaxInt + CorSz/2) % (CorSz) - CorSz/2 );

// Now zero is really zero. We just need rescaling now.
    dXmax = dxCorMaxInt/double(FineFactor) ;
    dYmax = dyCorMaxInt/double(FineFactor) ;

// Show correlation function.
/*    QImage* img = newQImageFromData (CorSz, CorSz, Correlation, corMax*0.97,corMax) ;
    QLabel* labs = new QLabel () ;
    labs->setPixmap (QPixmap::fromImage(*img)) ;
    labs->show ()  ;*/
}


void Correlator::DivideByRootOf (Correlator& divisor)
{
    if ( ! Correlation || ! divisor.Correlation )
    {
        cout << "Error : absent data." << endl ;
        return ;
    }

    double minimum = CorSz/2*CorSz/2*128 ;

    for (int idx=0 ; idx<CorSz*CorSz ; idx++)
    {
        if ( divisor.Correlation [idx][0] > minimum*minimum)
            Correlation [idx][0] /= sqrt(divisor.Correlation [idx][0]);
        else
            Correlation [idx][0] /= minimum;
    }
}


void Correlator::freeRefMemory (void)
{
    fftwMutex.lock() ;
    if (refGray)
       	fftw_free(refGray);
    refGray = 0 ;
    fftwMutex.unlock() ;
}

void Correlator::freeTargAndCorrelMemory (void)
{
    fftwMutex.lock() ;
    if (targGray)
        fftw_free(targGray); 
    targGray = 0 ;
    Correlation = 0 ;
    fftwMutex.unlock() ;
}

void Correlator::freeFourierMemory (void)
{
    fftwMutex.lock() ;
	if (fTargGray)
        fftw_free(fTargGray);
    fTargGray = 0 ;
    fftwMutex.unlock() ;

    fftwMutex.lock() ;
    if (fRefGray)
        fftw_free(fRefGray);
    fRefGray = 0 ;
    fftwMutex.unlock() ;
}

